--- Error Object
-- @module generic.err

-- Copyright (C) 2007-2016 emlix GmbH, see file AUTHORS
--
-- This file is part of e2factory, the emlix embedded build system.
-- For more information see http://www.e2factory.org
--
-- e2factory is a registered trademark of emlix GmbH.
--
-- e2factory is free software: you can redistribute it and/or modify it under
-- the terms of the GNU General Public License as published by the
-- Free Software Foundation, either version 3 of the License, or (at your
-- option) any later version.
--
-- This program is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
-- FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
-- more details.

local err =  {}
local e2lib = require("e2lib")
local strict = require("strict")

local function assert_err(e)
    assert(type(e) == "table", "No error table supplied: "..debug.traceback())
    assert(type(e.count) == "number", "Error count is not a number")
    assert(type(e.msg) == "table", "Error message table of wrong type")
    return true
end

--- append a string to an error object
-- @param format string: format string
-- @param ... list of strings required for the format string
-- @return table: the error object
function err.append(e, format, ...)
    assert_err(e)
    assert(type(format) == "string")
    e.count = e.count + 1
    table.insert(e.msg, string.format(format, ...))
    return e
end

--- insert an error object into another one
-- @param e table: the error object
-- @param re table: the error object to insert
-- @param ... list of strings required for the format string
-- @return table: the error object
function err.cat(e, re, ...)
    assert_err(e)
    assert(type(re) == "string" or assert_err(re))
    -- auto-convert strings to error objects before inserting
    if type(re) == "string" then
        if #{...} > 0 then
            re = err.new(re, ...)
        else
            re = err.new("%s", re)
        end
    end
    table.insert(e.msg, re)
    e.count = e.count + 1
    return e
end

--- print error messages at log level 1
-- @param e table: the error object
-- @param depth number: used internally to count and display nr. of sub errors
function err.print(e, depth)
    assert_err(e)
    assert(type(depth) == "number" or depth == nil)
    if not depth then
        depth = 1
    else
        depth = depth + 1
    end

    local prefix = string.format("Error [%d]: ", depth)

    for k,m in ipairs(e.msg) do
        if type(m) == "string" then
            e2lib.logf(1, "%s%s", prefix, m)
            prefix = string.format("      [%d]: ", depth)
        else
            -- it's a sub error
            m:print(depth)
        end
    end
end

--- set the error counter
-- @param e the error object
-- @param n number: new error counter setting
-- @return nil
function err.setcount(e, n)
    assert_err(e)
    assert(type(n) == "number")
    e.count = n
end

--- get the error counter
-- @param e the error object
-- @return number: the error counter
function err.getcount(e)
    assert_err(e)
    return e.count
end

--- create an error object
-- @param format string: format string
-- @param ... list of arguments required for the format string
-- @return table: the error object
function err.new(format, ...)
    assert(type(format) == "string" or format == nil)
    local e = {}
    e.count = 0
    e.msg = {}
    local meta = { __index = err }
    setmetatable(e, meta)
    if format then
        e:append(format, ...)
    end
    return e
end

return strict.lock(err)

-- vim:sw=4:sts=4:et:
